@@ -135,10 +135,14 @@ split_paths([], _S, Path, Paths) ->
 
 -spec call(term()) -> term().
 call(Req) ->
+    Ref = erlang:monitor(process, ?MODULE),
     ?MODULE ! {code_call, self(), Req},
     receive 
 	{?MODULE, Reply} ->
-	    Reply
+            erlang:demonitor(Ref,[flush]),
+	    Reply;
+        {'DOWN',Ref,process,_,_} ->
+            exit({'DOWN',code_server,Req})
     end.
 
 reply(Pid, Res) ->
@@ -807,7 +811,13 @@ clear_namedb([], _) ->
 %% Dir must be a complete pathname (not only a name).
 insert_dir(Dir, Db) ->
     Splitted = filename:split(Dir),
-    Name = get_name_from_splitted(Splitted),
+    case get_name_from_splitted(Splitted) of
+	Name when Name /= "ebin", Name /= "." ->
+	    Name;
+	_ ->
+	    SplittedAbsName = filename:split(absname(Dir)),
+	    Name = get_name_from_splitted(SplittedAbsName)
+    end,
     AppDir = filename:join(del_ebin_1(Splitted)),
     do_insert_name(Name, AppDir, Db).
 
@@ -933,15 +943,25 @@ del_ebin(Dir) ->
     filename:join(del_ebin_1(filename:split(Dir))).
 
 del_ebin_1([Parent,App,"ebin"]) ->
-    Ext = archive_extension(),
-    case filename:basename(Parent, Ext) of
-	Parent ->
-	    %% Plain directory.
+    case filename:basename(Parent) of
+	[] ->
+	    %% Parent is the root directory
 	    [Parent,App];
-	Archive ->
-	    %% Archive.
-	    [Archive]
+	_ ->
+	    Ext = archive_extension(),
+	    case filename:basename(Parent, Ext) of
+		Parent ->
+		    %% Plain directory.
+		    [Parent,App];
+		Archive ->
+		    %% Archive.
+		    [Archive]
+	    end
     end;
+del_ebin_1(Path = [_App,"ebin"]) ->
+    del_ebin_1(filename:split(absname(filename:join(Path))));
+del_ebin_1(["ebin"]) ->
+    del_ebin_1(filename:split(absname("ebin")));
 del_ebin_1([H|T]) ->
     [H|del_ebin_1(T)];
 del_ebin_1([]) ->
@@ -1110,19 +1130,18 @@ try_load_module_2(File, Mod, Bin, From, Architecture,
                   #state{moddb=Db}=St) ->
     case catch hipe_unified_loader:load_native_code(Mod, Bin, Architecture) of
         {module,Mod} = Module ->
-	    ets:insert(Db, [{{native,Mod},true},{Mod,File}]),
+	    ets:insert(Db, {Mod,File}),
             {reply,Module,St};
         no_native ->
             try_load_module_3(File, Mod, Bin, From, Architecture, St);
         Error ->
             error_msg("Native loading of ~ts failed: ~p\n", [File,Error]),
-            {reply,ok,St}
+            {reply,{error,Error},St}
     end.
 
-try_load_module_3(File, Mod, Bin, From, Architecture, St0) ->
+try_load_module_3(File, Mod, Bin, From, _Architecture, St0) ->
     Action = fun({module,_}=Module, #state{moddb=Db}=S) ->
 		     ets:insert(Db, {Mod,File}),
-		     post_beam_load([Mod], Architecture, S),
 		     {reply,Module,S};
 		({error,on_load_failure}=Error, S) ->
 		     {reply,Error,S};
@@ -1133,24 +1152,14 @@ try_load_module_3(File, Mod, Bin, From, Architecture, St0) ->
     Res = erlang:load_module(Mod, Bin),
     handle_on_load(Res, Action, Mod, From, St0).
 
-hipe_result_to_status(Result, #state{moddb=Db}) ->
+hipe_result_to_status(Result, #state{}) ->
     case Result of
-	{module,Mod} ->
-	    ets:insert(Db, [{{native,Mod},true}]),
+	{module,_} ->
 	    Result;
 	_ ->
 	    {error,Result}
     end.
 
-post_beam_load(_, undefined, _) ->
-    %% HiPE is disabled.
-    ok;
-post_beam_load(Mods0, _Architecture, #state{moddb=Db}) ->
-    %% post_beam_load/2 can potentially be very expensive because it
-    %% blocks multi-scheduling. Therefore, we only want to call
-    %% it with modules that are known to have native code loaded.
-    Mods = [M || M <- Mods0, ets:member(Db, {native,M})],
-    hipe_unified_loader:post_beam_load(Mods).
 
 int_list([H|T]) when is_integer(H) -> int_list(T);
 int_list([_|_])                    -> false;
@@ -1293,15 +1302,12 @@ abort_if_sticky(L, Db) ->
 	[_|_] -> {error,Sticky}
     end.
 
-do_finish_loading(Prepared, #state{moddb=Db}=St) ->
+do_finish_loading(Prepared, #state{moddb=Db}) ->
     MagicBins = [B || {_,{B,_}} <- Prepared],
     case erlang:finish_loading(MagicBins) of
 	ok ->
 	    MFs = [{M,F} || {M,{_,F}} <- Prepared],
 	    true = ets:insert(Db, MFs),
-	    Ms = [M || {M,_} <- MFs],
-	    Architecture = erlang:system_info(hipe_architecture),
-	    post_beam_load(Ms, Architecture, St),
 	    ok;
 	{Reason,Ms} ->
 	    {error,[{M,Reason} || M <- Ms]}
@@ -1372,11 +1378,10 @@ finish_on_load(PidRef, OnLoadRes, #state{on_load=OnLoad0}=St0) ->
 
 finish_on_load_1(Mod, OnLoadRes, Waiting, St) ->
     Keep = OnLoadRes =:= ok,
-    erlang:finish_after_on_load(Mod, Keep),
+    erts_code_purger:finish_after_on_load(Mod, Keep),
     Res = case Keep of
 	      false ->
 		  _ = finish_on_load_report(Mod, OnLoadRes),
-		  _ = erts_code_purger:purge(Mod),
 		  {error,on_load_failure};
 	      true ->
 		  {module,Mod}
@@ -1404,7 +1409,7 @@ finish_on_load_report(Mod, Term) ->
     %% from the code_server process.
     spawn(fun() ->
 		  F = "The on_load function for module "
-		      "~s returned ~P\n",
+		      "~s returned:~n~P\n",
 
 		  %% Express the call as an apply to simplify
 		  %% the ext_mod_dep/1 test case.
